# coding=utf-8

from pyage.core import address

from pyage.core.stop_condition import StepLimitStopCondition
from pyage_forams.solutions.environment import environment_factory, Environment3d
from pyage_forams.solutions.foram import create_forams, create_agent
from pyage_forams.solutions.genom import GenomFactory
from pyage_forams.solutions.insolation_meter import StaticInsolation, DynamicInsolation
from pyage_forams.solutions.statistics import SimpleStatistics, PsiStatistics


factory = GenomFactory(chambers_limit=5)
genom_factory = lambda: factory.generate
forams = create_forams(8, initial_energy=5)
agents = create_agent
insolation_meter = lambda: StaticInsolation(surface_insolation=5, insolation_factor=0.1)
#DynamicInsolation([(20, 10, 0.2), (10, 20, 0.4)])
size = lambda: 60

reproduction_minimum = lambda: 35
movement_energy = lambda: 0.15
growth_minimum = lambda: 15
energy_need = lambda: 0.5
algae_limit = lambda: 100000000
newborn_limit = lambda: 29
initial_algae_probability = lambda: 0.3

environment = environment_factory(regeneration_factor=0.5, clazz=Environment3d)

stop_condition = lambda: StepLimitStopCondition(290)

address_provider = address.SequenceAddressProvider
stats = SimpleStatistics


import logging
from random import randint, choice, random

from pyage.core.inject import Inject
from pyage_forams.solutions.cell import Cell


logger = logging.getLogger(__name__)


class AbstractEnvironment(object):
    @Inject("insolation_meter", "algae_limit", "initial_algae_probability")
    def __init__(self, regeneration_factor):
        super(AbstractEnvironment, self).__init__()
        self.regeneration_factor = regeneration_factor

    def add_foram(self, foram):
        choice(filter(lambda c: c.is_empty(), self.get_all_cells())).insert_foram(foram)

    def tick(self, step):
        for cell in self.get_all_cells():
            if 0 < cell.get_algae() < self.algae_limit:
                cell.add_algae(self.regeneration_factor + self.insolation_meter.get_insolation(cell, step))
        while random() > 0.7:
            try:
                choice(filter(lambda c: c.is_empty(), self.get_all_cells())).add_algae(random() * 2)
            except:
                pass

    def get_cell(self, address):
        for cell in self.get_all_cells():
            if cell.get_address() == address:
                return cell

    def add_neighbour(self, cell_address, neighbour):
        self.get_cell(cell_address).add_neighbour(neighbour)


class Environment2d(AbstractEnvironment):
    @Inject("size")
    def __init__(self, regeneration_factor):
        super(Environment2d, self).__init__(regeneration_factor)
        self.grid = self._initialize_grid()

    def _initialize_grid(self):
        grid = [[Cell(randint(0, 4) if random() < self.initial_algae_probability else 0) for _ in range(self.size)] for
                _ in range(self.size)]
        for i in range(self.size):
            for j in range(self.size):
                grid[i][j]._neighbours.extend([grid[x][y] for x in range(max(0, i - 1), min(self.size, i + 2)) for y in
                                               range(max(0, j - 1), min(self.size, j + 2)) if x != i or y != j])
        return grid

    def get_all_cells(self):
        for row in self.grid:
            for cell in row:
                yield cell

    def get_border_cells(self, side):
        if side == "right":
            return self._get_right_cells()
        elif side == "left":
            return self._get_left_cells()
        elif side == "upper":
            return self._get_upper_cells()
        elif side == "lower":
            return self._get_lower_cells()
        else:
            raise KeyError("unknown side: %s" % side)

    def get_shadows(self, side):
        return [cell.to_shadow() for cell in self.get_border_cells(side)]

    def _get_left_cells(self):
        return [row[0] for row in self.grid]

    def _get_right_cells(self):
        return [row[-1] for row in self.grid]

    def _get_upper_cells(self):
        return [c for c in self.grid[0]]

    def _get_lower_cells(self):
        return [c for c in self.grid[-1]]

    def join_cells(self, cells, side):
        _join_rows(cells, self.get_border_cells(side))
        return {cell.get_address(): cell for cell in cells}


class Environment3d(AbstractEnvironment):
    @Inject("size")
    def __init__(self, regeneration_factor):
        super(Environment3d, self).__init__(regeneration_factor)
        self.grid = self._initialize_grid()

    def _initialize_grid(self):
        grid = [[[Cell(randint(1, 4) if random() < self.initial_algae_probability else 0)
                  for _ in range(self.size)] for _ in range(self.size)] for _ in range(self.size)]
        for i in range(self.size):
            for j in range(self.size):
                for k in range(self.size):
                    grid[i][j][k].depth = self.size - k - 1
                    grid[i][j][k]._neighbours.extend(
                        [grid[x][y][z] for x in range(max(0, i - 1), min(self.size, i + 2)) for y in
                         range(max(0, j - 1), min(self.size, j + 2)) for z in
                         range(max(0, k - 1), min(self.size, k + 2)) if x != i or y != j or z != k])

        return grid

    def get_all_cells(self):
        for plane in self.grid:
            for row in plane:
                for cell in row:
                    yield cell

    def get_border_cells(self, side):
        if side == "right":
            return self._get_right_cells()
        elif side == "left":
            return self._get_left_cells()
        elif side == "upper":
            return self._get_upper_cells()
        elif side == "lower":
            return self._get_lower_cells()
        elif side == "front":
            return self._get_front_cells()
        elif side == "back":
            return self._get_back_cells()
        else:
            raise KeyError("unknown side: %s" % side)

    def get_shadows(self, side):
        return [[cell.to_shadow() for cell in row] for row in self.get_border_cells(side)]

    def _get_left_cells(self):
        return [row[0] for row in self.grid]

    def _get_right_cells(self):
        return [row[-1] for row in self.grid]

    def _get_upper_cells(self):
        return [c for c in self.grid[0]]

    def _get_lower_cells(self):
        return [c for c in self.grid[-1]]

    def _get_front_cells(self):
        return [[row[0] for row in plane] for plane in self.grid]

    def _get_back_cells(self):
        return [[row[-1] for row in plane] for plane in self.grid]

    def join_cells(self, cells, side):
        for (r1, r2) in zip(cells, self.get_border_cells(side)):
            _join_rows(r1, r2)
        for (r1, r2) in zip(cells[1:], self.get_border_cells(side)[:-1]):
            _join_rows(r1, r2)
        for (r1, r2) in zip(cells[:-1], self.get_border_cells(side)[1:]):
            _join_rows(r1, r2)

        return {cell.get_address(): cell for row in cells for cell in row}


def _join_rows(r1, r2):
    for (c1, c2) in zip(r1, r2):
        c1.add_neighbour(c2)
        c2.add_neighbour(c1)
    for (c1, c2) in zip(r1[1:], r2[:-1]):
        c1.add_neighbour(c2)
        c2.add_neighbour(c1)
    for (c1, c2) in zip(r1[:-1], r2[1:]):
        c1.add_neighbour(c2)
        c2.add_neighbour(c1)


def environment_factory(regeneration_factor=0.1, clazz=Environment2d):
    e = [None]

    def environ():
        if e[0] is None:
            e[0] = clazz(regeneration_factor)
        return e[0]

    return environ


import logging

from pyage_forams.solutions.distributed.request import TakeAlgaeRequest
from pyage.core.inject import Inject
from pyage_forams.solutions.distributed.requests.migrate import MigrateRequest


logger = logging.getLogger(__name__)


class ShadowCell(object):
    @Inject("request_dispatcher")
    def __init__(self, address, available_food, algae, empty,agent_address):
        self.food = available_food
        self._algae = algae
        self.address = address
        self.agent_address = agent_address
        self.empty = empty
        self._neighbours = []  # TODO get rid of this property

    def remove_foram(self):
        self.empty = True

    def insert_foram(self, foram):
        if hasattr(foram, "parent"):
            foram.parent.remove_foram(foram.get_address())
        else:
            logger.info("%s has no parent", foram)
        self.export_foram(foram)
        self.empty = False

    def take_algae(self, demand):
        to_let = min(demand, self._algae)
        self._algae -= to_let
        if not to_let == 0:
            self.request_dispatcher.submit_request(TakeAlgaeRequest(self.agent_address, self.address, to_let))
        return to_let

    def get_algae(self):
        return self._algae

    def add_algae(self, algae):
        self._algae += algae

    def is_empty(self):
        return self.empty

    def get_address(self):
        return self.address

    def available_food(self):
        return self.food

    def get_neighbours(self):
        return self._neighbours

    def add_neighbour(self, neighbour):
        self._neighbours.append(neighbour)

    def export_foram(self, foram):
        logger.info("exporting %s to %s" % (foram, self.address))
        self.request_dispatcher.submit_request(
            MigrateRequest(self.agent_address, self.address, foram))

    def update(self, (address, available_food, algae, empty, _)):
        logger.debug("%s  updating! %s", self, (address, available_food, algae, _))
        self._algae = algae
        self.food = available_food
        self.empty = empty

    def __repr__(self):
        return "(%d, ShadowCell, %d)" % (self._algae, self.available_food())





import logging
from random import random
from time import sleep, time

import Pyro4

from pyage.core.address import Addressable
from pyage.core.agent.agent import AGENT
from pyage.core.inject import Inject
from pyage_forams.solutions.distributed.neighbour_matcher import opposite


logger = logging.getLogger(__name__)


class RemoteForamAggregateAgent(Addressable):
    @Inject("forams", "environment", "neighbour_matcher", "request_dispatcher", "ns_hostname", "neighbours")
    def __init__(self, name=None):
        if name is not None:
            self.name = name
        super(RemoteForamAggregateAgent, self).__init__()
        self.updates = []
        self.requests = []
        self.joined = {}
        self._step = 0
        for foram in self.forams.values():
            foram.parent = self
            self.environment.add_foram(foram)
        self.neighbour_matcher.match_neighbours(self)

    def step(self):
        self._step += 1
        self._wait_for_neighbours()
        for foram in self.forams.values():
            if foram.cell is None or foram.cell.foram is None:
                logger.warning("something went wrong %s" % foram)
            foram.step()
        self.environment.tick(self._step)
        self._notify_neighbours()
        self.request_dispatcher.send_requests()
        self._process_requests()
        for update in self.updates:
            update()

    def stop(self):
        logger.info("finished")
        sleep(5)  # to let neighbours finish

    def get_steps(self):
        return self._step

    def _wait_for_neighbours(self):
        deadline = time() + 60
        while not self._all_neighbours_ready():
            if time() > deadline:
                logger.warning("timeout while waiting for neighbours")
                raise RuntimeError("Timeout in step %s waiting for neighbours: connected: %s, expected: %s" % (
                    self._step, self.joined, self.neighbours))
            logger.info("waiting for neighbours %d %s" % (self._step, self.joined))
            self._process_requests()
            sleep(random())  # TODO improve

    def _process_requests(self):
        while self.requests:
            self.requests.pop().execute(self)

    def _all_neighbours_ready(self):
        if len(self.joined) < len(self.neighbours):
            self.neighbour_matcher.match_neighbours(self)
        if len(self.joined) < len(self.neighbours):
            sleep(2)
            return False
        for (neighbour, step) in self.joined.iteritems():
            if step < self._step - 1:  # TODO make configurable
                return False
        return True

    def _notify_neighbours(self):
        try:
            for (neighbour, step) in self.joined.iteritems():
                self._notify_neighbour(neighbour)
        except:
            logger.exception("could not notify neighbours")

    def remove_foram(self, address):
        logger.info("removing foram %s", address)
        foram = self.forams[address]
        del self.forams[address]
        foram.parent = None
        return foram

    def add_foram(self, foram):
        foram.parent = self
        self.forams[foram.get_address()] = foram

    def take_algae(self, cell_address, algae):
        logger.debug("taking %f algae from %s" % (algae, cell_address))
        self.environment.get_cell(cell_address).take_algae(algae)

    def get_cells(self, side):
        return self.environment.get_border_cells(side)

    def get_shadows(self, side):
        return self.environment.get_shadows(side)

    def join(self, remote_address, shadow_cells, side, step):
        if not remote_address.startswith(AGENT + "." + self.neighbours[side]):
            logger.warning("Received request to %s-join with %s, expected neighbour on this side is %s" % (
                side, remote_address, self.neighbours[side]))
            return
        mapping = self.environment.join_cells(shadow_cells, side)

        def update():
            try:
                self.joined[remote_address] = self.neighbour_matcher.update(remote_address, side, mapping)
            except:
                logging.exception("could not update")

        self.updates.append(update)
        self.joined[remote_address] = step
        logger.info("%s is now %s-joined with: %s" % (self.address, side, remote_address))

    def import_foram(self, cell_address, foram):
        try:
            logger.warning("importing %s into %s" % (foram, cell_address))
            self.environment.get_cell(cell_address).insert_foram(foram)
            self.add_foram(foram)
            logger.warning("foram imported")
        except:
            logging.exception("Could not import cell")

    def submit_requests(self, requests):
        self.requests.extend(requests)

    def notify(self, neighbour_address, steps):
        logger.info("received notification %s %s", steps, neighbour_address)
        self.joined[neighbour_address] = steps

    def _notify_neighbour(self, neighbour):
        try:
            ns = Pyro4.locateNS(self.ns_hostname)
            agent = Pyro4.Proxy(ns.lookup(neighbour))
            agent.notify(AGENT + "." + self.get_address(), self.get_steps())
            logger.info("notified %s", neighbour)
        except:
            logger.exception("could not notify neighbour")


def create_remote_agent(name):
    agent = RemoteForamAggregateAgent(name)
    return {agent.get_address(): agent}


from unittest import TestCase

from pyage.core import inject
import pyage_forams.conf.dummy_conf
from pyage_forams.solutions.cell import Cell

from pyage_forams.solutions.environment import Environment2d, Environment3d


class TestDistributed(TestCase):
    @classmethod
    def setUpClass(cls):
        super(TestDistributed, cls).setUpClass()
        inject.config = "pyage_forams.conf.dummy_conf"

    def test_2d_joining(self):
        environment = Environment2d(0.2)
        cells = [Cell() for _ in range(3)]
        environment.join_cells(cells, "right")

        self.assertTrue(cells[0] in environment.grid[0][2].get_neighbours())
        self.assertTrue(cells[1] in environment.grid[0][2].get_neighbours())

        self.assertTrue(cells[0] in environment.grid[1][2].get_neighbours())
        self.assertTrue(cells[1] in environment.grid[1][2].get_neighbours())
        self.assertTrue(cells[2] in environment.grid[1][2].get_neighbours())

        self.assertTrue(cells[1] in environment.grid[2][2].get_neighbours())
        self.assertTrue(cells[2] in environment.grid[2][2].get_neighbours())

    def test_3d_joining(self):
        environment = Environment3d(0.2)
        cells = [[Cell() for _ in range(3)] for _ in range(3)]
        environment.join_cells(cells, "right")

        self.assertTrue(cells[0][0] in environment.grid[0][2][0].get_neighbours())
        self.assertTrue(cells[0][1] in environment.grid[0][2][0].get_neighbours())
        self.assertTrue(cells[1][0] in environment.grid[0][2][0].get_neighbours())
        self.assertTrue(cells[1][1] in environment.grid[0][2][0].get_neighbours())

        self.assertTrue(cells[0][0] in environment.grid[0][2][1].get_neighbours())
        self.assertTrue(cells[0][1] in environment.grid[0][2][1].get_neighbours())
        self.assertTrue(cells[0][2] in environment.grid[0][2][1].get_neighbours())
        self.assertTrue(cells[1][0] in environment.grid[0][2][1].get_neighbours())
        self.assertTrue(cells[1][1] in environment.grid[0][2][1].get_neighbours())
        self.assertTrue(cells[1][2] in environment.grid[0][2][1].get_neighbours())

        self.assertTrue(cells[0][1] in environment.grid[0][2][2].get_neighbours())
        self.assertTrue(cells[0][2] in environment.grid[0][2][2].get_neighbours())
        self.assertTrue(cells[1][1] in environment.grid[0][2][2].get_neighbours())
        self.assertTrue(cells[1][2] in environment.grid[0][2][2].get_neighbours())

        self.assertTrue(cells[0][0] in environment.grid[1][2][0].get_neighbours())
        self.assertTrue(cells[0][1] in environment.grid[1][2][0].get_neighbours())
        self.assertTrue(cells[1][0] in environment.grid[1][2][0].get_neighbours())
        self.assertTrue(cells[1][1] in environment.grid[1][2][0].get_neighbours())
        self.assertTrue(cells[2][0] in environment.grid[1][2][0].get_neighbours())
        self.assertTrue(cells[2][1] in environment.grid[1][2][0].get_neighbours())

        self.assertTrue(cells[0][0] in environment.grid[1][2][1].get_neighbours())
        self.assertTrue(cells[0][1] in environment.grid[1][2][1].get_neighbours())
        self.assertTrue(cells[0][2] in environment.grid[1][2][1].get_neighbours())
        self.assertTrue(cells[1][0] in environment.grid[1][2][1].get_neighbours())
        self.assertTrue(cells[1][1] in environment.grid[1][2][1].get_neighbours())
        self.assertTrue(cells[1][2] in environment.grid[1][2][1].get_neighbours())
        self.assertTrue(cells[2][0] in environment.grid[1][2][1].get_neighbours())
        self.assertTrue(cells[2][1] in environment.grid[1][2][1].get_neighbours())
        self.assertTrue(cells[2][2] in environment.grid[1][2][1].get_neighbours())

        self.assertTrue(cells[0][2] in environment.grid[1][2][2].get_neighbours())
        self.assertTrue(cells[0][1] in environment.grid[1][2][2].get_neighbours())
        self.assertTrue(cells[1][2] in environment.grid[1][2][2].get_neighbours())
        self.assertTrue(cells[1][1] in environment.grid[1][2][2].get_neighbours())
        self.assertTrue(cells[2][2] in environment.grid[1][2][2].get_neighbours())
        self.assertTrue(cells[2][1] in environment.grid[1][2][2].get_neighbours())

        self.assertTrue(cells[2][0] in environment.grid[2][2][0].get_neighbours())
        self.assertTrue(cells[2][1] in environment.grid[2][2][0].get_neighbours())
        self.assertTrue(cells[1][0] in environment.grid[2][2][0].get_neighbours())
        self.assertTrue(cells[1][1] in environment.grid[2][2][0].get_neighbours())

        self.assertTrue(cells[2][0] in environment.grid[2][2][1].get_neighbours())
        self.assertTrue(cells[2][1] in environment.grid[2][2][1].get_neighbours())
        self.assertTrue(cells[2][2] in environment.grid[2][2][1].get_neighbours())
        self.assertTrue(cells[1][0] in environment.grid[2][2][1].get_neighbours())
        self.assertTrue(cells[1][1] in environment.grid[2][2][1].get_neighbours())
        self.assertTrue(cells[1][2] in environment.grid[2][2][1].get_neighbours())

        self.assertTrue(cells[2][2] in environment.grid[2][2][2].get_neighbours())
        self.assertTrue(cells[2][1] in environment.grid[2][2][2].get_neighbours())
        self.assertTrue(cells[1][2] in environment.grid[2][2][2].get_neighbours())
        self.assertTrue(cells[1][1] in environment.grid[2][2][2].get_neighbours())



import logging

from pyage.core.address import Addressable


logger = logging.getLogger(__name__)


class Cell(Addressable):
    def __init__(self, algae=0):
        super(Cell, self).__init__()
        self._algae = algae
        self.foram = None
        self._neighbours = []

    def insert_foram(self, foram):
        logger.info("inserting %s into %s" % (foram, self))
        if not self.is_empty():
            raise ValueError("cannot insert foram to already occupied cell")
        self.foram = foram
        foram.cell = self

    def remove_foram(self):
        foram = self.foram
        foram.cell = None
        self.foram = None
        return foram

    def is_empty(self):
        return self.foram is None

    def take_algae(self, demand):
        to_let = min(demand, self._algae)
        self._algae -= to_let
        return to_let

    def get_algae(self):
        return self._algae

    def add_algae(self, algae):
        self._algae += algae

    def available_food(self):
        return self._algae + sum(map(lambda c: c.get_algae(), self._neighbours))

    def add_neighbour(self, cell):
        self._neighbours.append(cell)

    def get_neighbours(self):
        return self._neighbours

    def to_shadow(self):
        return self.get_address(), self.available_food(), self._algae, self.is_empty(), [cell.get_address() for cell in
                                                                                         self.get_neighbours()]

    def __repr__(self):
        return "(%d, %s, %d)" % (self._algae, self.foram, self.available_food())


from collections import defaultdict
import logging

import Pyro4

from pyage.core.agent.agent import AGENT
from pyage.core.inject import Inject
from pyage_forams.solutions.agent.shadow_cell import ShadowCell
from pyage_forams.solutions.distributed.request import Request


logger = logging.getLogger(__name__)


class NeighbourMatcher(object):
    @Inject("request_dispatcher", "size", "neighbours", 'ns_hostname')
    def __init__(self):
        super(NeighbourMatcher, self).__init__()
        self._located_agents = defaultdict(self._locate_agent)

    def match_neighbours(self, agent):
        for (side, address) in self.neighbours.iteritems():
            neighbour = self._locate_neighbour(address)
            if neighbour:
                self._join(neighbour, agent, side)

    def _locate_neighbour(self, address):
        try:
            ns = Pyro4.locateNS(self.ns_hostname)
            agents = ns.list(AGENT + "." + address)
            return Pyro4.Proxy(agents.values().pop())
        except:
            logging.warning("could not locate %s" % address)

    def _join(self, neighbour, agent, side):
        raise NotImplementedError()

    def _locate_agent(self, remote_address):
        ns = Pyro4.locateNS(self.ns_hostname)
        agent = Pyro4.Proxy(ns.lookup(remote_address))
        return agent


class Neighbour2dMatcher(NeighbourMatcher):
    def _join(self, remote_agent, agent, side):
        try:
            remote_address = AGENT + "." + remote_agent.get_address()
            logger.info("%s matching with: %s" % (side, remote_address))
            shadows = remote_agent.get_shadows(opposite(side))
            logger.debug("received shadows: %s" % shadows)
            shadow_cells = self.create_shadow_cells(remote_address, shadows)
            agent.join(remote_address, shadow_cells, side, remote_agent.get_steps())
            self.request_dispatcher.submit_request(
                Match2dRequest(remote_address, agent.get_shadows(side),
                               AGENT + "." + agent.get_address(), opposite(side), agent.get_steps()))
        except Exception, e:
            logger.exception("could not join: %s", e.message)

    @staticmethod
    def create_shadow_cells(remote_address, shadows):
        shadow_cells = {address: ShadowCell(address, available_food, algae, empty, remote_address) for
                        (address, available_food, algae, empty, _) in shadows}
        for (address, _, _, _, neighbours) in shadows:
            for neighbour in neighbours:
                try:
                    shadow_cells[address].add_neighbour(shadow_cells[neighbour])
                except KeyError:
                    pass
        return shadow_cells.values()

    def update(self, remote_address, side, mapping):
        logger.info("updating shadow cels from: %s" % remote_address)
        agent = self._locate_agent(remote_address)
        shadows = agent.get_shadows(opposite(side))
        for shadow in shadows:
            if shadow[0] in mapping:
                mapping[shadow[0]].update(shadow)
            else:
                logger.info("unsuccessful attempt to update cell with address %s", shadow.get_address())
        return agent.get_steps()


class Neighbour3dMatcher(NeighbourMatcher):
    def _join(self, remote_agent, agent, side):
        try:
            remote_address = AGENT + "." + remote_agent.get_address()
            logger.info("%s matching with: %s" % (side, remote_address))
            shadows = remote_agent.get_shadows(opposite(side))
            logger.debug("received shadows: %s" % shadows)
            shadow_cells = self.create_shadow_cells(remote_address, shadows)
            agent.join(remote_address, shadow_cells, side, remote_agent.get_steps())
            self.request_dispatcher.submit_request(
                Match3dRequest(remote_address, agent.environment.get_shadows(side),
                               AGENT + "." + agent.get_address(), opposite(side), agent.get_steps()))
        except Exception, e:
            logger.exception("could not join: %s", e.message)


    @staticmethod
    def create_shadow_cells(remote_address, shadows):
        shadow_cells = [[ShadowCell(address, available_food, algae, empty, remote_address) for
                         (address, available_food, algae, empty, _) in row] for row in shadows]
        mapping = {shadow_cell.get_address(): shadow_cell for row in shadow_cells for shadow_cell in row}
        for row in shadows:
            for (address, _, _, _, neighbours) in row:
                for neighbour in neighbours:
                    try:
                        mapping[address].add_neighbour(mapping[neighbour])
                    except KeyError:
                        pass
        return shadow_cells

    def update(self, remote_address, side, mapping):
        logger.info("updating shadow cels from: %s" % remote_address)
        agent = self._locate_agent(remote_address)
        shadows = agent.get_shadows(opposite(side))
        for row in shadows:
            for shadow in row:
                if shadow[0] in mapping:
                    mapping[shadow[0]].update(shadow)
                else:
                    logger.info("unsuccessful attempt to update cell with address %s", shadow.get_address())
        return agent.get_steps()


class Match2dRequest(Request):
    def __init__(self, agent_address, cells, remote_address, side, steps):
        super(Match2dRequest, self).__init__(agent_address)
        self.side = side
        self.shadows = cells
        self.remote_address = remote_address
        self.steps = steps

    def execute(self, agent):
        logger.info("2d-matching with %s" % self.remote_address)
        shadow_cells = Neighbour2dMatcher.create_shadow_cells(self.remote_address, self.shadows)
        agent.join(self.remote_address, shadow_cells, self.side, self.steps)


class Match3dRequest(Request):
    def __init__(self, agent_address, shadows, remote_address, side, steps):
        super(Match3dRequest, self).__init__(agent_address)
        self.side = side
        self.shadows = shadows
        self.remote_address = remote_address
        self.steps = steps

    def execute(self, agent):
        logger.info("3d-matching with %s" % self.remote_address)
        shadow_cells = Neighbour3dMatcher.create_shadow_cells(self.remote_address, self.shadows)
        agent.join(self.remote_address, shadow_cells, self.side, self.steps)


def opposite(side):
    if side == "left":
        return "right"
    elif side == "right":
        return "left"
    elif side == "upper":
        return "lower"
    elif side == "lower":
        return "upper"
    elif side == "front":
        return "back"
    elif side == "back":
        return "front"
    else:
        raise ValueError("unrecognized side: " + side)



import logging

from pyage.core.address import Addressable
from pyage.core.inject import Inject


logger = logging.getLogger(__name__)


class ForamAggregateAgent(Addressable):
    @Inject("forams", "environment")
    def __init__(self):
        super(ForamAggregateAgent, self).__init__()
        for foram in self.forams.values():
            foram.parent = self
            self.environment.add_foram(foram)

    def step(self):
        for foram in self.forams.values():
            foram.step()
        self.environment.tick(self.parent.steps)

    def remove_foram(self, address):
        foram = self.forams[address]
        del self.forams[address]
        foram.parent = None
        return foram

    def add_foram(self, foram):
        foram.parent = self
        self.forams[foram.get_address()] = foram


from pyage_forams.solutions.distributed.request import Request


class MigrateRequest(Request):
    def __init__(self, agent_address, cell_address, foram):
        super(MigrateRequest, self).__init__(agent_address)
        self.cell_address = cell_address
        self.foram = foram

    def execute(self, agent):
        agent.import_foram(self.cell_address, self.foram)

    def __repr__(self):
        return "MigrateRequest[%s, %s, %s]" % (self.agent_address, self.cell_address, self.foram)


from collections import defaultdict
import logging

import Pyro4

from pyage.core.inject import Inject


logger = logging.getLogger(__name__)


class Request(object):
    def __init__(self, agent_address):
        super(Request, self).__init__()
        self.agent_address = agent_address

    def execute(self, agent):
        raise NotImplementedError()


class TakeAlgaeRequest(Request):
    def __init__(self, agent_address, cell_address, algae):
        super(TakeAlgaeRequest, self).__init__(agent_address)
        self.algae = algae
        self.cell_address = cell_address

    def execute(self, agent):
        if not self.algae == 0:
            agent.take_algae(self.cell_address, self.algae)


class RequestDispatcher(object):
    @Inject('ns_hostname')
    def __init__(self):
        self.requests = defaultdict(list)

    def submit_request(self, request):
        self.requests[request.agent_address].append(request)
        logger.info("request submited %s" % request)

    def send_requests(self):
        try:
            logger.info("sending %s" % self.requests)
            ns = Pyro4.locateNS(self.ns_hostname)
            for (agent_address, requests) in self.requests.iteritems():
                if requests:
                    logger.info("sending %s to %s" % ( requests, agent_address))
                    agent = Pyro4.Proxy(ns.lookup(agent_address))
                    agent.submit_requests(requests)
                    requests[:] = []
        except:
            logging.exception("could not send request")


def create_dispatcher():
    d = [None]

    def environ():
        if d[0] is None:
            d[0] = RequestDispatcher()
        return d[0]

    return environ


def _pickle_method(method):
    func_name = method.im_func.__name__
    obj = method.im_self
    cls = method.im_class
    if func_name.startswith('__') and not func_name.endswith('__'):
        cls_name = cls.__name__.lstrip('_')
        if cls_name: func_name = '_' + cls_name + func_name
    return _unpickle_method, (func_name, obj, cls)


def _unpickle_method(func_name, obj, cls):
    for cls in cls.mro():
        try:
            func = cls.__dict__[func_name]
        except KeyError:
            pass
        else:
            break
    return func.__get__(obj, cls)


import copy_reg
import types

copy_reg.pickle(types.MethodType, _pickle_method, _unpickle_method)



# coding=utf-8
from functools import partial

import Pyro4
from pyage.core import address

from pyage.core.stop_condition import StepLimitStopCondition
from pyage_forams.solutions.distributed.neighbour_matcher import Neighbour3dMatcher
from pyage_forams.solutions.agent.remote_aggegate import create_remote_agent
from pyage_forams.solutions.distributed.request import create_dispatcher
from pyage_forams.solutions.environment import environment_factory, Environment3d
from pyage_forams.solutions.foram import create_forams
from pyage_forams.solutions.genom import GenomFactory
from pyage_forams.solutions.insolation_meter import StaticInsolation
from pyage_forams.solutions.statistics import PlottingStatistics, SimpleStatistics


factory = GenomFactory(chambers_limit=5)
genom_factory = lambda: factory.generate
forams = create_forams(1, initial_energy=5)
agents = partial(create_remote_agent, "lowerleftfront")
insolation_meter = StaticInsolation
size = lambda: 5

environment = environment_factory(regeneration_factor=0.1, clazz=Environment3d)
neighbour_matcher = Neighbour3dMatcher

request_dispatcher = create_dispatcher()

stop_condition = lambda: StepLimitStopCondition(90)

reproduction_minimum = lambda: 10
movement_energy = lambda: 0.25
growth_minimum = lambda: 10
energy_need = lambda: 0.2
algae_limit = lambda: 2
newborn_limit = lambda: 9
reproduction_probability = lambda: 0.8
growth_probability = lambda: 0.8
growth_cost_factor = lambda: 0.5
capacity_factor = lambda: 1.1
initial_algae_probability = lambda: 0.03

address_provider = address.SequenceAddressProvider
stats = SimpleStatistics

ns_hostname = lambda: "127.0.0.1"
pyro_daemon = Pyro4.Daemon()
daemon = lambda: pyro_daemon

neighbours = lambda: {"right": "lowerrightfront", "upper": "upperleftfront", "back": "lowerleftback"}